package gc;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
/**
 * <p>Title: Actor Memory</p>
 * <p>Description: This class simulates the memory and its operations</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: WWC</p>
 * @author WeiJen Wang
 * @version 1.0
 */

public class ActorMemory implements java.io.Serializable{

  //it indicates an inherent root actor, such as StandardOutput
  private boolean root=false;

  //snapshot indicates that this actor is currently in GC
  transient private boolean snapshot=false;

  //localMark is the mark bit used in a mark-sweep algorithm
  transient private boolean localMark=false;
  transient private boolean pseudoRoot=false;

  //The following two variables are used to distinguish non-local-reachable actors and
  //local pseudo root reachable actors
  //Only non-local reachable actors are possible part of global garbage
  transient private boolean nonLocalMark=false;
  transient private boolean nonLocalPseudoRoot=false;

  //active indicates that an actor is processing or has messages in its message box.
  transient private boolean active=false;
  transient private boolean activeAlternative;

  //fList stores information of its forward acquaintances and their related information
  private ForwardList fList;

  //iList stores information of its inverse acquaintances, and their related information
  private InverseList iList;

  //locks a actor for waiting an incoming message.
  private int lockCounter=0;
  transient private boolean islocked=false;

  transient private String globalMark=null;

  public ActorMemory(String selfRef) {
    fList=new ForwardList(selfRef);
    iList=new InverseList(selfRef);
    lockCounter=0;
  }

  //mark and sweep utility
  public synchronized void resetLocalMark() {localMark=false;}
  public synchronized void colorLocalMark() {localMark=true;}
  public synchronized boolean isLocalMarked() {return localMark || pseudoRoot;}

  public synchronized void resetPseudoRoot() {pseudoRoot=false;}
  public synchronized void colorPseudoRoot() {pseudoRoot=true;}
  public synchronized boolean isPseudoRoot() {return pseudoRoot;}


  //global GC functions for a local collector
  public synchronized void resetNonLocalMark() {this.nonLocalMark=false;}
  public synchronized void colorNonLocalMark() {this.nonLocalMark=true;}
  public synchronized boolean isNonLocalMarked() {return this.nonLocalMark;}

  public synchronized void resetNonLocalPseudoRoot() {nonLocalPseudoRoot=false;}
  public synchronized void colorNonLocalPseudoRoot() {nonLocalPseudoRoot=true;}
  public synchronized boolean isNonLocalPseudoRoot() {return nonLocalPseudoRoot;}
  public synchronized boolean isMarked() {return localMark || pseudoRoot || nonLocalMark || nonLocalPseudoRoot;}


  public synchronized void resetGlobalMark() {globalMark=null;}
  public synchronized void colorGlobalMark(String GMark) {globalMark=GMark;}
  public synchronized String isGlobalMarked() {return globalMark;}


  /*lock operation functions: incLock(), decLock(), isLocked(), and getLock()
   *
   */
  public synchronized void incLock() {
//System.out.println("INC LOCK:" + this.fList.getSelfRef());
    if (snapshot) {islocked=true;}
    lockCounter++;
  }

  public synchronized void decLock() {
//System.out.println("DEC LOCK:" + this.fList.getSelfRef());
    lockCounter--;
  }
  public synchronized boolean isLocked() {
    if (snapshot) {return islocked;}
    else {
      return lockCounter != 0;
    }
  }

  public synchronized int getLock() {return lockCounter;}

  public synchronized ForwardList GCGetForwardList() {return fList;}
  public synchronized InverseList GCGetinverseList() {return iList;}

  public synchronized ForwardList getForwardList() {return fList;}

  public synchronized InverseList getInverseList() {
    if (snapshot) {
      try {
        this.islocked=true;
//System.err.println("...AM.Waiting: " + fList.getSelfRef());
        wait();
//System.err.println("...AM..Awaken: " + fList.getSelfRef());
      }
      catch (Exception e) {
        System.err.println("GC error at Class ActorMemory, method getInverseList():"+e);
      }
    }
    return iList;
  }


  /***During GC:
   *  1. An actor can only become unblocked, but cannot become blocked
   *  2. val is stored in the variable activeAlternative
   */
  public synchronized void setActive(boolean val) {
    if (snapshot) {
      if (val==true) {
        active = true;
        activeAlternative=true;
      } else {
        activeAlternative=false;
      }
    } else {active=val;}
  }

  public  synchronized boolean isActive() { return active;}

  /***
   * If GC ends, set active=activeAlternative to restore its real value
   * If GC starts, set activeAlternative=active
   */
  public  void startGC() {
    activeAlternative=active;
    snapshot=true;
    islocked=(lockCounter!=0);
    resetLocalMark();
    resetPseudoRoot();
    resetNonLocalMark();
    resetNonLocalPseudoRoot();
    fList.startGC();
  }

  public synchronized void endGC() {
    active=activeAlternative;
    snapshot=false;
    fList.endGC();
    notifyAll();
  }

  public  synchronized boolean getSnapshotBit() {return snapshot;}

  public synchronized String toString() {
    String res="[== " + fList.getSelfRef() + " ==],lock="+isLocked()+", active="+isActive()+"\n";;
    res+=fList.toString()+"\n------------\n"+this.iList.toString();
    return res+"\n[-------------------]\n";
  }

  private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException{
    out.defaultWriteObject();
    out.flush();
  }

  private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, java.lang.ClassNotFoundException{
    in.defaultReadObject();
    localMark=false;
    active=false;
    islocked=false;
  }
}